/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "gui.h"
#include "config_gui.h"
#include <glib-object.h>

#define CONFIG_GUI_TYPE_OBJECT       (config_gui_get_type ())
#define CONFIG_GUI(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), CONFIG_GUI_TYPE_OBJECT, ConfigGUI))
#define CONFIG_GUI_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CONFIG_GUI_TYPE_OBJECT, ConfigGUIClass))
#define IS_CONFIG_GUI(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), CONFIG_GUI_TYPE_OBJECT))
#define IS_CONFIG_GUI_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CONFIG_GUI_TYPE_OBJECT))
#define CONFIG_GUI_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), CONFIG_GUI_TYPE_OBJECT, ConfigGUIClass))

typedef struct ConfigGUIClass ConfigGUIClass;

ConfigGUI *config_gui_get_object(void);
GType config_gui_get_type (void);

struct ConfigGUI
{
    GObject parent;

    GStaticRWLock rwlock;
    ConfigGUIOpts opts;
};

struct ConfigGUIClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };
static ConfigGUI *configuration = NULL;

static void config_gui_load(ConfigGUI *config);
static void config_gui_save(ConfigGUI *config);

G_DEFINE_TYPE(ConfigGUI, config_gui, G_TYPE_OBJECT)

static void
config_gui_init (ConfigGUI *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
    memset(&obj->opts, 0, sizeof(ConfigGUIOpts));
}

static void
config_gui_finalize (GObject *obj)
{
    ConfigGUI *config = CONFIG_GUI(obj);

    config_gui_save(config);

    g_static_rw_lock_free(&config->rwlock);
}

static void
config_gui_class_init (ConfigGUIClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = config_gui_finalize;

    signals[CONFIG_CHANGED_SIGNAL] = g_signal_new  ("config_changed",
                                                    G_OBJECT_CLASS_TYPE (klass),
                                                    G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                    0,
                                                    NULL, NULL,
                                                    g_cclosure_marshal_VOID__INT,
                                                    G_TYPE_NONE, 1, G_TYPE_INT);
}

static ConfigGUI *
config_gui_new(void)
{
    ConfigGUI *config = g_object_new(CONFIG_GUI_TYPE_OBJECT, NULL);

    config_gui_load(config);

    return config;
}

ConfigGUIOpts
config_gui_get_opts(void)
{
    ConfigGUIOpts opts;

    g_assert(configuration != NULL);

    g_static_rw_lock_reader_lock(&configuration->rwlock);
    opts = configuration->opts;
    g_static_rw_lock_reader_unlock(&configuration->rwlock);

    return opts;
}

void
config_gui_set_opts(ConfigGUIOpts new_opts)
{
    g_assert(configuration != NULL);

    g_static_rw_lock_writer_lock(&configuration->rwlock);
    configuration->opts = new_opts;
    g_static_rw_lock_writer_unlock(&configuration->rwlock);

    config_gui_save(configuration);

    g_signal_emit(configuration, signals[CONFIG_CHANGED_SIGNAL], 0);
}

static void
config_gui_load(ConfigGUI *config)
{
    GKeyFile *keyfile;
    GError *err = NULL;
    char *filename;
    char **groups;
    int i;
    char *tmp;

    g_assert(config != NULL);

    g_static_rw_lock_writer_lock(&config->rwlock);

    filename = g_build_filename(g_get_user_config_dir(), "NNTPGrab", "nntpgrab_gui.conf", NULL);
    keyfile = g_key_file_new();
    if (!g_key_file_load_from_file(keyfile, filename, G_KEY_FILE_NONE, &err)) {
        // No config file yet, create a new one
        memset(&config->opts, 0, sizeof(config->opts));

        g_static_rw_lock_writer_unlock(&config->rwlock);

#if 0
// Ignore errors for now
        GtkWidget *windowMain;
        char *errmsg;

        windowMain = get_widget("windowMain");
        errmsg = g_strdup_printf(_("config_gui_load(): Error while opening file '%s'\n%s"), filename, err->message);
        dialog_show(windowMain, errmsg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

        g_free(errmsg);
#endif
        g_free(filename);
        g_error_free(err);
        g_key_file_free(keyfile);

        // Set default values
        strncpy(config->opts.hostname1, "localhost", sizeof(config->opts.hostname1));
        config->opts.show_column_progress = TRUE;
        config->opts.show_column_subject = TRUE;
        config->opts.show_column_poster = TRUE;
        config->opts.show_column_newsgroups = TRUE;
        config->opts.show_column_filesize = TRUE;
        config->opts.show_column_filesize_remaining = TRUE;
        config->opts.show_column_stamp = TRUE;
        config->opts.show_column_time_remaining = TRUE;
        config->opts.show_column_time_to_finish = TRUE;
        config->opts.enable_auto_import = FALSE;
#ifdef WIN32
        tmp = g_build_path(G_DIR_SEPARATOR_S, g_get_user_special_dir(G_USER_DIRECTORY_DOCUMENTS), "NNTPGrab", "NZB", NULL);
#else
        tmp = g_build_path(G_DIR_SEPARATOR_S, g_get_home_dir(), "NNTPGrab", "NZB", NULL);
#endif
        strncpy(config->opts.auto_import_directory, tmp, sizeof(config->opts.auto_import_directory));
        g_mkdir_with_parents(tmp, 0700);
        g_free(tmp);

        config_gui_save(config);

        return;
    }

    g_free(filename);

    groups = g_key_file_get_groups(keyfile, NULL);
    i = 0;
    while (groups[i]) {
        if (!strcmp(groups[i], "options")) {
            config->opts.show_notification_on_file_downloaded = g_key_file_get_boolean(keyfile, groups[i], "show_notification_on_file_downloaded", NULL);
            config->opts.show_notification_on_file_decoded = g_key_file_get_boolean(keyfile, groups[i], "show_notification_on_file_decoded", NULL);
            config->opts.show_tray_icon = g_key_file_get_boolean(keyfile, groups[i], "show_tray_icon", NULL);

            tmp = g_key_file_get_string(keyfile, groups[i], "hostname1", NULL);
            if (tmp && strlen(tmp) > 0) {
                strncpy(config->opts.hostname1, tmp, sizeof(config->opts.hostname1));
                g_free(tmp);
            } else {
                // Use the default value on a fresh configuration
                strncpy(config->opts.hostname1, "localhost", sizeof(config->opts.hostname1));
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "hostname2", NULL);
            if (tmp) {
                strncpy(config->opts.hostname2, tmp, sizeof(config->opts.hostname2));
                g_free(tmp);
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "hostname3", NULL);
            if (tmp) {
                strncpy(config->opts.hostname3, tmp, sizeof(config->opts.hostname3));
                g_free(tmp);
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "hostname4", NULL);
            if (tmp) {
                strncpy(config->opts.hostname4, tmp, sizeof(config->opts.hostname4));
                g_free(tmp);
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "hostname5", NULL);
            if (tmp) {
                strncpy(config->opts.hostname5, tmp, sizeof(config->opts.hostname5));
                g_free(tmp);
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_progress", NULL)) {
                config->opts.show_column_progress = g_key_file_get_boolean(keyfile, groups[i], "show_column_progress", NULL);
            } else {
                config->opts.show_column_progress = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_subject", NULL)) {
                config->opts.show_column_subject = g_key_file_get_boolean(keyfile, groups[i], "show_column_subject", NULL);
            } else {
                config->opts.show_column_subject = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_poster", NULL)) {
                config->opts.show_column_poster = g_key_file_get_boolean(keyfile, groups[i], "show_column_poster", NULL);
            } else {
                config->opts.show_column_poster = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_newsgroups", NULL)) {
                config->opts.show_column_newsgroups = g_key_file_get_boolean(keyfile, groups[i], "show_column_newsgroups", NULL);
            } else {
                config->opts.show_column_newsgroups = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_filesize", NULL)) {
                config->opts.show_column_filesize = g_key_file_get_boolean(keyfile, groups[i], "show_column_filesize", NULL);
            } else {
                config->opts.show_column_filesize = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_filesize_remaining", NULL)) {
                config->opts.show_column_filesize_remaining = g_key_file_get_boolean(keyfile, groups[i], "show_column_filesize_remaining", NULL);
            } else {
                config->opts.show_column_filesize_remaining = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_stamp", NULL)) {
                config->opts.show_column_stamp = g_key_file_get_boolean(keyfile, groups[i], "show_column_stamp", NULL);
            } else {
                config->opts.show_column_stamp = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_time_remaining", NULL)) {
                config->opts.show_column_time_remaining = g_key_file_get_boolean(keyfile, groups[i], "show_column_time_remaining", NULL);
            } else {
                config->opts.show_column_time_remaining = TRUE;
            }

            if (g_key_file_has_key(keyfile, groups[i], "show_column_time_to_finish", NULL)) {
                config->opts.show_column_time_to_finish = g_key_file_get_boolean(keyfile, groups[i], "show_column_time_to_finish", NULL);
            } else {
                config->opts.show_column_time_to_finish = TRUE;
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "auto_import_directory", NULL);
            if (tmp) {
                strncpy(config->opts.auto_import_directory, tmp, sizeof(config->opts.auto_import_directory));
                g_free(tmp);
            }

            if (strlen(config->opts.auto_import_directory) == 0) {
#ifdef WIN32
                tmp = g_build_path(G_DIR_SEPARATOR_S, g_get_user_special_dir(G_USER_DIRECTORY_DOCUMENTS), "NNTPGrab", "NZB", NULL);
#else
                tmp = g_build_path(G_DIR_SEPARATOR_S, g_get_home_dir(), "NNTPGrab", "NZB", NULL);
#endif
                strncpy(config->opts.auto_import_directory, tmp, sizeof(config->opts.auto_import_directory));
                g_mkdir_with_parents(tmp, 0700);
                g_free(tmp);
            }

            if (g_key_file_has_key(keyfile, groups[i], "enable_auto_import", NULL)) {
                config->opts.enable_auto_import = g_key_file_get_boolean(keyfile, groups[i], "enable_auto_import", NULL);
            } else {
                config->opts.enable_auto_import = FALSE;
            }

            tmp = g_key_file_get_string(keyfile, groups[i], "last_used_nzb_directory", NULL);
            if (tmp && strlen(tmp) > 0) {
                strncpy(config->opts.last_used_nzb_directory, tmp, sizeof(config->opts.last_used_nzb_directory));
                g_free(tmp);
            } else {
                // No default necessary
            }

            i++;
            continue;
        }

        i++;
    }

    g_strfreev(groups);

    g_key_file_free(keyfile);

    g_static_rw_lock_writer_unlock(&config->rwlock);
}

static void
config_gui_save(ConfigGUI *config)
{
    GKeyFile *keyfile;
    GError *err = NULL;
    char *filename;
    char *dirname;
    char *contents;

    g_assert(config != NULL);

    g_static_rw_lock_writer_lock(&config->rwlock);

    keyfile = g_key_file_new();

    g_key_file_set_boolean(keyfile, "options", "show_notification_on_file_downloaded", config->opts.show_notification_on_file_downloaded);
    g_key_file_set_boolean(keyfile, "options", "show_notification_on_file_decoded", config->opts.show_notification_on_file_decoded);
    g_key_file_set_boolean(keyfile, "options", "show_tray_icon", config->opts.show_tray_icon);
    g_key_file_set_string (keyfile, "options", "hostname1", config->opts.hostname1);
    g_key_file_set_string (keyfile, "options", "hostname2", config->opts.hostname2);
    g_key_file_set_string (keyfile, "options", "hostname3", config->opts.hostname3);
    g_key_file_set_string (keyfile, "options", "hostname4", config->opts.hostname4);
    g_key_file_set_string (keyfile, "options", "hostname5", config->opts.hostname5);
    g_key_file_set_boolean(keyfile, "options", "show_column_progress", config->opts.show_column_progress);
    g_key_file_set_boolean(keyfile, "options", "show_column_subject", config->opts.show_column_subject);
    g_key_file_set_boolean(keyfile, "options", "show_column_poster", config->opts.show_column_poster);
    g_key_file_set_boolean(keyfile, "options", "show_column_newsgroups", config->opts.show_column_newsgroups);
    g_key_file_set_boolean(keyfile, "options", "show_column_filesize", config->opts.show_column_filesize);
    g_key_file_set_boolean(keyfile, "options", "show_column_filesize_remaining", config->opts.show_column_filesize_remaining);
    g_key_file_set_boolean(keyfile, "options", "show_column_stamp", config->opts.show_column_stamp);
    g_key_file_set_boolean(keyfile, "options", "show_column_time_remaining", config->opts.show_column_time_remaining);
    g_key_file_set_boolean(keyfile, "options", "show_column_time_to_finish", config->opts.show_column_time_to_finish);
    g_key_file_set_string (keyfile, "options", "auto_import_directory", config->opts.auto_import_directory);
    g_key_file_set_boolean(keyfile, "options", "enable_auto_import", config->opts.enable_auto_import);
    g_key_file_set_string (keyfile, "options", "last_used_nzb_directory", config->opts.last_used_nzb_directory);

    contents = g_key_file_to_data(keyfile, NULL, &err);
    if (!contents) {
        GtkWidget *windowMain;
        char *errmsg;

        windowMain = nntpgrab_gui_base_get_widget("windowMain");
        errmsg = g_strdup_printf(_("config_gui_save(): Error while saving config file\n%s"), err->message);
        nntpgrab_gui_base_dialog_show(windowMain, errmsg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

        g_free(errmsg);
        g_error_free(err);
        g_free(contents);
        g_key_file_free(keyfile);

        g_static_rw_lock_writer_unlock(&config->rwlock);

        return;
    }

    // Create the folder ~/.config/NNTPGrab if it didn't exist already
    dirname = g_build_path(G_DIR_SEPARATOR_S, g_get_user_config_dir(), "NNTPGrab", NULL);
    g_mkdir_with_parents(dirname, 0700);
    g_free(dirname);

    filename = g_build_filename(g_get_user_config_dir(), "NNTPGrab", "nntpgrab_gui.conf", NULL);

    if (!g_file_set_contents(filename, contents, -1, &err)) {
        GtkWidget *windowMain;
        char *errmsg;

        windowMain = nntpgrab_gui_base_get_widget("windowMain");
        errmsg = g_strdup_printf(_("config_gui_save(): Error while opening file '%s'\n%s"), filename, err->message);
        nntpgrab_gui_base_dialog_show(windowMain, errmsg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

        g_free(errmsg);
        g_error_free(err);
        g_free(contents);
        g_free(filename);
        g_key_file_free(keyfile);

        g_static_rw_lock_writer_unlock(&config->rwlock);

        return;
    }

    g_free(contents);
    g_free(filename);
    g_key_file_free(keyfile);

    g_static_rw_lock_writer_unlock(&config->rwlock);
}

void
config_gui_initialize(void)
{
    configuration = config_gui_new();
}

ConfigGUI *
config_gui_get_object(void)
{
    return configuration;
}
