/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "nntpgrab_gui_base.h"
#include "icon_64x64.xpm"
#if !GTK_CHECK_VERSION(2,12,0)
#include <glade/glade.h>
#endif

#if GTK_CHECK_VERSION(2,12,0)
static GtkBuilder *builder = NULL;
#else
static GladeXML *xml = NULL;
#endif

#if GTK_CHECK_VERSION(2,12,0)
GObject *
nntpgrab_gui_base_get_object(const char *name)
{
    GObject *obj;

    if (!builder) {
        return NULL;
    }

    obj = gtk_builder_get_object(builder, name);
    if (!obj) {
        g_critical(_("No widget found with name '%s'\n"), name);
    }

    return obj;
}

GtkWidget *
nntpgrab_gui_base_get_widget(const char *name)
{
    return GTK_WIDGET(nntpgrab_gui_base_get_object(name));
}

#else

GObject *
nntpgrab_gui_base_get_object(const char *name)
{
    // Glade doesn't really know about objects..
    return G_OBJECT(nntpgrab_gui_base_get_widget(name));
}

GtkWidget *
nntpgrab_gui_base_get_widget(const char *name)
{
    GtkWidget *widget;

    if (!xml) {
        return NULL;
    }

    widget = glade_xml_get_widget(xml, name);
    if (!widget) {
        g_critical(_("No widget found with name '%s'\n"), name);
    }

    return widget;
}
#endif

static gboolean
window_state_changed(GtkWidget *caller, GdkEvent *event, gpointer data)
{
    if (!nntpgrab_gui_base_tray_icon_get_is_shown()) {
        return FALSE;
    }

#if GTK_CHECK_VERSION(2,10,0)
    if (    (event->window_state.new_window_state & GDK_WINDOW_STATE_ICONIFIED) == GDK_WINDOW_STATE_ICONIFIED &&
            (event->window_state.changed_mask & GDK_WINDOW_STATE_ICONIFIED) == GDK_WINDOW_STATE_ICONIFIED) {
#ifdef WIN32
        gtk_widget_hide(caller);
#else
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(caller), TRUE);
#endif
    }
#endif

    return FALSE;
}

typedef struct _log_func_data {
    char msg[1024];
    GLogLevelFlags log_level;
} LogFuncData;

static gboolean
log_func_phase2(gpointer data)
{
    LogFuncData *func_data = (LogFuncData*) data;
    GtkWidget *window;

    window = nntpgrab_gui_base_get_widget("windowMain");

    nntpgrab_gui_base_dialog_show(window, func_data->msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

    if (func_data->log_level & G_LOG_LEVEL_ERROR) {
        gtk_main_quit();
    }

    g_slice_free(LogFuncData, func_data);

    return FALSE;
}

static void
log_func (const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer user_data)
{
    LogFuncData *func_data = g_slice_new(LogFuncData);
    char *msg;

    if (log_domain) {
        msg = g_strdup_printf("%s: %s", log_domain, message);
    } else {
        msg = g_strdup(message);
    }

    strncpy(func_data->msg, msg, sizeof(func_data->msg) - 1);
    func_data->log_level = log_level;

    g_print("%s\n", msg);

    g_idle_add(log_func_phase2, func_data);

    g_free(msg);
}

void
nntpgrab_gui_base_create_ui(const char *interface_name)
{
    GtkWidget *windowMain;
    char *filename;
    GdkPixbuf *pixbuf;
    GError *err = NULL;
    const char *path;
#ifdef WIN32
    char *dirname;
    char *dirname2;
#endif
    char *interface_filename;

#ifdef WIN32
    dirname = g_win32_get_package_installation_directory_of_module(NULL);
    dirname2 = g_build_path(G_DIR_SEPARATOR_S, dirname, "\\share\\locale", NULL);
    bindtextdomain (GETTEXT_PACKAGE, dirname2);
    g_free(dirname2);
#else
    bindtextdomain (GETTEXT_PACKAGE, NNTPGRABLOCALEDIR);
#endif
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    g_log_set_handler(NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab-utils", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab-automation", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab-server", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);

    /* load the interface */
#ifdef WIN32
    path = g_build_path(G_DIR_SEPARATOR_S, dirname, "\\share\\nntpgrab", NULL);
    g_free(dirname);
#else
    if (g_getenv("NNTPGRAB_DATA_DIR")) {
        path = g_getenv("NNTPGRAB_DATA_DIR");
    } else {
        path = SHARE_DIR;
    }
#endif

#if GTK_CHECK_VERSION(2,12,0)
    interface_filename = g_strdup_printf("%s.ui", interface_name);
#else
    interface_filename = g_strdup_printf("%s.glade", interface_name);
#endif

    filename = g_build_filename(path, G_DIR_SEPARATOR_S, interface_filename, NULL);

    g_free(interface_filename);

#if GTK_CHECK_VERSION(2,12,0)
    builder = gtk_builder_new();
    if (!gtk_builder_add_from_file(builder, filename, &err)) {
        builder = NULL;
        g_error(err->message);
    }

    gtk_builder_connect_signals(builder, NULL);
#else
    xml = glade_xml_new(filename, NULL, NULL);

    /* connect the signals in the interface */
    glade_xml_signal_autoconnect(xml);
#endif

    g_free(filename);

    // Set the icon of the window
    windowMain = nntpgrab_gui_base_get_widget("windowMain");
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) &icon_64x64_xpm);
    gtk_window_set_icon(GTK_WINDOW(windowMain), pixbuf);
    g_object_unref(pixbuf);

    g_signal_connect(G_OBJECT(windowMain), "window_state_event",
                           G_CALLBACK(window_state_changed), NULL);
}
