/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib-object.h>
#include <gmodule.h>

#include "nntpconnection.h"
#include "nntpgrab_plugin_nntp.h"
#include "plugin_nntp.h"
#include "nntpgrab_internal.h"

typedef struct NNTPConnectionClass NNTPConnectionClass;

static PluginNNTPExportedFuncs plugin_funcs;

static GType nntpconnection_get_type (void);

static void update_progress_callback(int bytes_downloaded, gpointer obj);
static void update_traffic_monitor (int bytes_received[10], time_t stamp);
static void download_begin_callback (gpointer obj);
static void connecting_callback(int conn_id, gpointer obj);
static void connected_callback(const char *welcome_msg, gpointer obj);
static void disconnect_callback(NNTPDisconnectType disconnect_type, const char *errmsg, gpointer obj);
static void parse_header_callback (int position, const char *subject, const char *poster, time_t postdate, const char *message_id, int size, int numlines, int start_range, int end_range, gpointer data);

struct NNTPConnection
{
    GObject parent;

    GStaticRWLock rwlock;
    Configuration *config;
    int num_servers;
    int conn_id;
    ConfigServer server;
    const char *collection_name;
    const char *subject;
    char *newsgroup;
    const NNTPPart *part;
};

struct NNTPConnectionClass
{
    GObjectClass parent;
};

G_DEFINE_TYPE(NNTPConnection, nntpconnection, G_TYPE_OBJECT)

static void
nntpconnection_init (NNTPConnection *obj)
{
    g_static_rw_lock_init(&obj->rwlock);

    obj->conn_id = -1;
    obj->collection_name = NULL;
    obj->subject = NULL;
    obj->part = NULL;
    obj->newsgroup = NULL;
    obj->config = NULL;
    obj->num_servers = 0;

    memset(&obj->server, 0, sizeof(obj->server));
}

static void
nntpconnection_finalize (GObject *obj)
{
    NNTPConnection *connection = NNTPCONNECTION(obj);

    g_static_rw_lock_writer_lock(&connection->rwlock);

    if (connection->newsgroup) {
        g_free(connection->newsgroup);
        connection->newsgroup = NULL;
    }

    g_static_rw_lock_writer_unlock(&connection->rwlock);

    g_object_unref(connection->config);

    g_static_rw_lock_free(&connection->rwlock);
}

static void
nntpconnection_class_init (NNTPConnectionClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = nntpconnection_finalize;
}

gboolean
nntpconnection_initialize(char **errmsg)
{
    GModule *plugin = NULL;
    PluginNNTPImportedFuncs imported_funcs;
    plugin_nntp_init_func init_func;
    char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#define EXT ".so"
#endif

    if (g_getenv("NNTPGRAB_LIBDIR")) {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_nntp%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
    } else {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_nntp%s", PLUGIN_DIR, EXT);
    }

    plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);
    g_free(filename);

    if (!plugin) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    imported_funcs.version = NNTPGRAB_PLUGIN_API_VERSION;
    imported_funcs.update_progress = update_progress_callback;
    imported_funcs.notify_download_begin = download_begin_callback;
    imported_funcs.notify_connecting = connecting_callback;
    imported_funcs.notify_connected = connected_callback;
    imported_funcs.notify_disconnect = disconnect_callback;
    imported_funcs.parse_header = parse_header_callback;
    imported_funcs.update_traffic_monitor = update_traffic_monitor;

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_get_version", (gpointer *) &plugin_funcs.get_version)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    // Check if the API version matches
    if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("NNTP Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_initialize", (gpointer *) &init_func)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_get_is_connected", (gpointer *) &plugin_funcs.get_is_connected)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_connect", (gpointer *) &plugin_funcs.connect)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_disconnect", (gpointer *) &plugin_funcs.disconnect)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_get_article_by_message_id", (gpointer *) &plugin_funcs.get_article_by_message_id)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_group", (gpointer *) &plugin_funcs.group)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_nntp_xover", (gpointer *) &plugin_funcs.xover)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    /* All required methods are found, lets start */
    if (!init_func(imported_funcs)) {
        if (errmsg) {
            *errmsg = g_strdup(_("NNTP Plugin initialization failed!"));
        }

        return FALSE;
    }

    nntpgrab_core_emit_debug_message(_("NNTP Plugin initialized"));

    return TRUE;
}

NNTPConnection *
nntpconnection_new(Configuration *config, ConfigServer server)
{
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    NNTPConnection *connection;
    GList *list;

    g_static_mutex_lock(&mutex);
    connection = g_object_new(NNTPCONNECTION_TYPE_OBJECT, NULL);
    g_static_mutex_unlock(&mutex);

    connection->server = server;
    connection->config = config;
    list = configuration_get_avail_servers(config);
    connection->num_servers = g_list_length(list);
    configuration_free_avail_servers(config,list);
    g_object_ref(config);

    return connection;
}

void
nntpconnection_destroy(NNTPConnection *obj)
{
    g_object_unref(obj);
}

NNTPGrabErrCode
nntpconnection_connect(NNTPConnection *obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);
    NNTPGrabErrCode retval;
    char *errmsg = NULL;
    char *msg;

    g_assert(plugin_funcs.connect);

    msg = g_strdup_printf(_("Initiating connection to %s (%s:%i)"), conn->server.servername, conn->server.hostname, conn->server.port);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    retval = plugin_funcs.connect(conn->server.hostname, conn->server.port, conn->server.username, conn->server.password, conn->server.use_ssl, &errmsg, &conn->conn_id, conn);

    msg = g_strdup_printf(_("Connect call returned %i"), retval);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    if (retval != NNTP_ERROR_NONE) {
        switch (retval) {
            case NNTP_ERROR_HOST_NOT_FOUND:
            case NNTP_ERROR_CONNECTION_REFUSED:
            case NNTP_ERROR_CONNECTION_TIMEOUT:
            case NNTP_ERROR_READ:
            case NNTP_ERROR_WRITE:
            case NNTP_ERROR_SSL_INITIALISE:
                // Ignore these, these are already processed by the connect function itself
                break;
            case NNTP_ERROR_INVALID_MSG:
                disconnect_callback(DISCONNECT_INVALID_MSG, errmsg, obj);
                break;
            case NNTP_ERROR_LOGIN_FAILURE:
                disconnect_callback(DISCONNECT_LOGIN_FAILURE, errmsg, obj);
                break;
            case NNTP_ERROR_TOO_MANY_CONNECTIONS:
                disconnect_callback(DISCONNECT_LOGIN_FAILURE, errmsg, obj);
                break;
            default:
                disconnect_callback(DISCONNECT_UNEXPECTED, NULL, obj);
        }

        if (errmsg) {
            g_free(errmsg);
        }
    }

    return retval;
}

void
nntpconnection_disconnect(NNTPConnection *obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);
    char *msg;

    g_assert(plugin_funcs.disconnect);

    msg = g_strdup_printf(_("Closing connection to %s (%s:%i)"), conn->server.servername, conn->server.hostname, conn->server.port);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    plugin_funcs.disconnect(conn->conn_id, conn);
}

gboolean
nntpconnection_get_is_connected(NNTPConnection *obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_assert(plugin_funcs.get_is_connected);

    return plugin_funcs.get_is_connected(conn->conn_id);
}

NNTPGrabErrCode
nntpconnection_get_part(NNTPConnection *obj, const char *collection_name, const char *subject, const NNTPPart *part, const char *filename)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);
    NNTPGrabErrCode ret;
    char *msg;

    g_assert(part);
    g_assert(plugin_funcs.get_article_by_message_id);

    msg = g_strdup_printf(_("Trying to retrieve part '%s' from subject '%s'"), part->message_id, subject);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    g_static_rw_lock_writer_lock(&conn->rwlock);
    conn->collection_name = collection_name;
    conn->subject = subject;
    conn->part = part;
    g_static_rw_lock_writer_unlock(&conn->rwlock);

    ret = plugin_funcs.get_article_by_message_id(conn->conn_id, part->message_id, NULL, filename, conn);

    msg = g_strdup_printf(_("Retrieve part function returned %i"), ret);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    g_static_rw_lock_writer_lock(&conn->rwlock);
    conn->part = NULL;
    conn->subject = NULL;
    conn->collection_name = NULL;
    g_static_rw_lock_writer_unlock(&conn->rwlock);

    return ret;
}

gboolean
nntpconnection_group(NNTPConnection *obj, int conn_id, const char *newsgroup)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);
    NNTPGrabErrCode ret;
    char *msg;
    ConfigGroupInfo info;

    g_assert(newsgroup);
    g_assert(plugin_funcs.group);

    msg = g_strdup_printf(_("Trying to select newsgroup '%s'"), newsgroup);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    g_static_rw_lock_writer_lock(&conn->rwlock);
    if (conn->newsgroup) {
        g_free(conn->newsgroup);
    }
    conn->newsgroup = g_strdup(newsgroup);
    g_static_rw_lock_writer_unlock(&conn->rwlock);

    info = configuration_get_newsgroup_info(conn->config, conn->server.servername, conn->newsgroup);
    ret = plugin_funcs.group(conn->conn_id, newsgroup, &info.num_articles, &info.start_article, &info.end_article, conn);

    msg = g_strdup_printf(_("GROUP function returned %i"), ret);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    if (ret == NNTP_ERROR_NONE) {
        // Save the num_articles, start_article and end_article in the configuration
        configuration_set_newsgroup_info(conn->config, conn->server.servername, newsgroup, info);
    }

    return (ret == NNTP_ERROR_NONE);
}

gboolean
nntpconnection_xover(NNTPConnection *obj, gboolean full_update)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);
    NNTPGrabErrCode ret;
    char *msg;
    ConfigGroupInfo info;

    g_assert(plugin_funcs.xover);

    msg = g_strdup_printf(_("Trying to retrieve headers using XOVER"));
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    g_assert (conn->newsgroup);
    g_static_rw_lock_reader_unlock(&conn->rwlock);

    info = configuration_get_newsgroup_info(conn->config, conn->server.servername, conn->newsgroup);

    if (full_update) {
        ret = plugin_funcs.xover(conn->conn_id, info.start_article, -1, conn);
    } else {
        ret = plugin_funcs.xover(conn->conn_id, 1, info.end_article, conn);
    }

    msg = g_strdup_printf(_("XOVER function returned %i"), ret);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    return (ret == NNTP_ERROR_NONE);
}

int
nntpconnection_get_conn_id(NNTPConnection *obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    return conn->conn_id;
}

/**
 * The NNTP plugin calls this function periodic to notify the progress of the download of the requested part
 *
 * @arg     bytes_downloaded    The number of bytes that are already downloaded
 * @arg     obj                 Pointer to internal NNTPGrab Core data
 */
static void
update_progress_callback(int bytes_downloaded, gpointer obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    nntpgrab_core_emit_part_progress_update(conn->server.servername, conn->conn_id, conn->subject, conn->part->partnum, bytes_downloaded, conn->part->size);
    g_static_rw_lock_reader_unlock(&conn->rwlock);
}

/**
 * The NNTP plugin calls this function every second to notify the used bandwidth
 *
 * @arg     bytes_received      A array containing the number of bytes transferred for the last 10 seconds
 * @arg     stamp               Timestamp of the latest measure
 */
static void
update_traffic_monitor(int bytes_received[10], time_t stamp)
{
    nntpgrab_core_emit_traffic_monitor_update(bytes_received, stamp);
}

/**
 * The NNTP plugin calls this function when the download of the requested part is about to start
 *
 * @arg     obj             Pointer to internal NNTPGrab Core data
 */
static void
download_begin_callback (gpointer obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    nntpgrab_core_emit_part_download_start(conn->server.servername, conn->conn_id, conn->collection_name, conn->subject, conn->part->partnum);
    g_static_rw_lock_reader_unlock(&conn->rwlock);
}

/**
 * The NNTP plugin calls this function when a connection is being set up
 *
 * @arg     obj             Pointer to internal NNTPGrab Core data
 */
static void connecting_callback(int conn_id, gpointer obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    nntpgrab_core_emit_connecting(conn->server.servername, conn_id);
    g_static_rw_lock_reader_unlock(&conn->rwlock);
}

/**
 * The NNTP plugin calls this function when the connection was set up successfully
 *
 * @arg     welcome_msg     The welcome message the usenet server sent us
 * @arg     obj             Pointer to internal NNTPGrab Core data
 */
static void connected_callback(const char *welcome_msg, gpointer obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    nntpgrab_core_emit_connected(conn->server.servername, conn->conn_id, welcome_msg);
    g_static_rw_lock_reader_unlock(&conn->rwlock);
}

/**
 * The NNTP plugin calls this function when a (possibly unexpected) disconnect has occured
 *
 * @arg     obj             Pointer to internal NNTPGrab Core data
 * @arg     type            The type of the disconnect
 * @arg     errmsg          An optional error message
 */
static void
disconnect_callback(NNTPDisconnectType disconnect_type, const char *errmsg, gpointer obj)
{
    NNTPConnection *conn = NNTPCONNECTION(obj);

    g_static_rw_lock_reader_lock(&conn->rwlock);
    nntpgrab_core_emit_disconnect(conn->server.servername, conn->conn_id, disconnect_type, errmsg);
    g_static_rw_lock_reader_unlock(&conn->rwlock);
}

static void
parse_header_callback (int position, const char *subject, const char *poster, time_t postdate, const char *message_id, int size, int numlines, int start_range, int end_range, gpointer data)
{

}
