/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <glib-object.h>

#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"
#include "par2.h"
#include "plugin_par2.h"
#include "nntpgrab_plugin_par2.h"

static PluginPar2ExportedFuncs plugin_funcs;

static void emit_repair_message_callback(const char *message);

typedef struct Par2Class Par2Class;

static GType par2_get_type (void);

struct Par2
{
    GObject parent;

    GStaticRWLock rwlock;
};

struct Par2Class
{
    GObjectClass parent;
};

G_DEFINE_TYPE(Par2, par2, G_TYPE_OBJECT);

static void
par2_init (Par2 *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
}

static void
par2_finalize (GObject *obj)
{
    Par2 *par2 = PAR2(obj);

    g_static_rw_lock_free(&par2->rwlock);
}

static void
par2_class_init (Par2Class *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = par2_finalize;
}

gboolean
par2_initialize(char **errmsg)
{
    GModule *plugin = NULL;
    PluginPar2ImportedFuncs imported_funcs;
    plugin_par2_init_func init_func;
    char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#define EXT ".so"
#endif

    if (g_getenv("NNTPGRAB_LIBDIR")) {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_par2%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
    } else {
        filename = g_strdup_printf("%s/libnntpgrab_plugin_par2%s", PLUGIN_DIR, EXT);
    }

    plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);

    g_free(filename);

    if (!plugin) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    imported_funcs.version = NNTPGRAB_PLUGIN_API_VERSION;
    imported_funcs.emit_repair_message = emit_repair_message_callback;

    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_get_version", (gpointer *) &plugin_funcs.get_version)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
        if (errmsg) {
            *errmsg = g_strdup_printf(_("PAR2 Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
        }

        return FALSE;
    }

    /* All required methods are found, lets start */
    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_initialize", (gpointer *) &init_func)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!g_module_symbol(plugin, "nntpgrab_plugin_par2_repair_files", (gpointer *) &plugin_funcs.repair_files)) {
        if (errmsg) {
            *errmsg = g_strdup(g_module_error());
        }

        return FALSE;
    }

    if (!init_func(imported_funcs)) {
        if (errmsg) {
            *errmsg = g_strdup(_("PAR2 Plugin initialization failed!"));
        }

        return FALSE;
    }

    nntpgrab_core_emit_debug_message(_("PAR2 plugin initialized"));

    return TRUE;
}

Par2 *
par2_new(void)
{
    Par2 *par2;

    par2 = g_object_new(PAR2_TYPE_OBJECT, NULL);

    return par2;
}

void
par2_destroy(Par2 *obj)
{
    g_object_unref(obj);
}

gboolean
par2_repair_files(Par2 *obj, const char *directory, const char *par2filename)
{
    gboolean retval;

    g_assert(plugin_funcs.repair_files);

    retval = plugin_funcs.repair_files(directory, par2filename);

    return retval;
}

static void
emit_repair_message_callback(const char *message)
{
    char *ptr = g_strdup(message);
    char *msg_stripped = ptr;

    // strip out any newlines
    while (*ptr) {
        if (*ptr == '\r') {
            *ptr = ' ';
        }
        if (*ptr == '\n') {
            *ptr = ' ';
        }
        ptr++;
    }

    nntpgrab_core_emit_par2_repair_message(msg_stripped);

    g_free(msg_stripped);
}
