/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef HAVE_NSS_COMPAT_OSSL
#include <nss_compat_ossl/nss_compat_ossl.h>
#else
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#ifdef _MSC_VER
#include <io.h>
#else
#include <sys/time.h>
#endif

#include "strptime.h"
#include "nntpgrab_plugin_nntp.h"

#ifdef _MSC_VER
#define S_IRUSR _S_IREAD
#define S_IWUSR _S_IWRITE

#undef  gai_strerror
#define gai_strerror gai_strerrorA

int gettimeofday (struct timeval *tv, void* tz)
{
  union {
    long long ns100; /*time since 1 Jan 1601 in 100ns units */
    FILETIME ft;
  } now;

  GetSystemTimeAsFileTime (&now.ft);
  tv->tv_usec = (long) ((now.ns100 / 10LL) % 1000000LL);
  tv->tv_sec = (long) ((now.ns100 - 116444736000000000LL) / 10000000LL);
  return (0);
}

#endif

static PluginNNTPImportedFuncs imported_funcs;

// Keep track of the number of bytes received for the last 10 seconds
static int bytes_received[10];
static GThread *traffic_thread = NULL;
static gboolean abort_traffic_thread = FALSE;
static GStaticRWLock traffic_rwlock = G_STATIC_RW_LOCK_INIT;

// Link a socket_id to a SSL pointer
static GHashTable *connections = NULL;
static GStaticRWLock connections_rwlock = G_STATIC_RW_LOCK_INIT;

static void free_connection_func(gpointer data);

typedef struct _nntp_connection_data {
    int socket_id;
    gboolean use_ssl;
    SSL *ssl;
    char recv_buf[4096];      // It should be safe to assume that one line is less than 4096 characters
    int recv_buf_len;
} NNTPConnectionData;

#ifdef WIN32
#define	timersub(tvp, uvp, vvp)						\
	do {								\
		(vvp)->tv_sec = (tvp)->tv_sec - (uvp)->tv_sec;		\
		(vvp)->tv_usec = (tvp)->tv_usec - (uvp)->tv_usec;	\
		if ((vvp)->tv_usec < 0) {				\
			(vvp)->tv_sec--;				\
			(vvp)->tv_usec += 1000000;			\
		}							\
	} while (0)
#endif

gpointer
traffic_thread_func(gpointer data)
{
    int empty_buf[10];
    gboolean null_buf_sent = FALSE;
    time_t timestamp = 0;
#if 0
    int i;
#endif

    memset(empty_buf, 0, sizeof(empty_buf));

    while (!abort_traffic_thread) {
        time_t prev_timestamp;

        // Traffic monitoring
        prev_timestamp = timestamp;
        timestamp = time(NULL);

        if (timestamp > prev_timestamp) {
            time_t diff;
            int bytes_received_copy[10];

            // Calculate how many steps we need to shift
            diff = timestamp - prev_timestamp;
            if (diff > 10) {
                diff = 10;
            }

            g_static_rw_lock_writer_lock(&traffic_rwlock);
            memcpy(bytes_received_copy, bytes_received, sizeof(bytes_received));

            // Shift all the previous values by 'diff'
            memmove(bytes_received, &bytes_received[diff], (10 - diff) * sizeof(int));
            memset(&bytes_received[10 - diff], 0, (10 - diff) * sizeof(int));

            g_static_rw_lock_writer_unlock(&traffic_rwlock);

            // Notify upstream about our values
            // If there hasn't been any sane data, send a empty buffer 1 time
            if (!memcmp(bytes_received_copy, empty_buf, sizeof(empty_buf))) {
                if (!null_buf_sent) {
                    null_buf_sent = TRUE;
                    imported_funcs.update_traffic_monitor(bytes_received_copy, timestamp);
                }
            } else {
                imported_funcs.update_traffic_monitor(bytes_received_copy, timestamp);
                null_buf_sent = FALSE;
            }

#if 0
            g_print("stamp = %li\n", timestamp);
            for (i = 0; i < 10; i++) {
                g_print("bytes_received[%i] = %i\n", i, bytes_received[i]);
            }
            g_print("\n");
#endif
        }

        g_usleep(G_USEC_PER_SEC * 1);
    }

    return NULL;
}

gboolean
nntpgrab_plugin_nntp_initialize(PluginNNTPImportedFuncs funcs)
{
    imported_funcs = funcs;

    memset(bytes_received, 0, sizeof(bytes_received));
    abort_traffic_thread = FALSE;                                                       // TODO: implement a nice clean up
    traffic_thread = g_thread_create(traffic_thread_func, NULL, TRUE, NULL);
    connections = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, free_connection_func);   // TODO: implement a nice clean up

    return TRUE;
}

int
nntpgrab_plugin_nntp_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

gboolean
nntpgrab_plugin_nntp_get_is_connected(int conn_id)
{

    if (conn_id <= 0) {
        return FALSE;
    } else {
        return TRUE;
    }
}

static int
get_status_code(const char *line)
{
    /* The status code is mentioned at the first 3 characters of the line.
     * atoi() doesn't detect errors and quits parsing after the first invalid
     * character which is enough for us */
    if (!line) {
        return -1;
    }

    return atoi(line);
}

static void
perform_traffic_monitoring(int bytes_read, gpointer data)
{
    g_static_rw_lock_writer_lock(&traffic_rwlock);
    bytes_received[9] += bytes_read;
    g_static_rw_lock_writer_unlock(&traffic_rwlock);
}

static void
free_connection_func(gpointer data)
{
    g_slice_free(NNTPConnectionData, data);
}

static void
add_connection(int socket_id, gboolean use_ssl, SSL *ssl)
{
    NNTPConnectionData *data = g_slice_new0(NNTPConnectionData);

    data->socket_id = socket_id;
    data->use_ssl = use_ssl;
    data->ssl = ssl;

    g_static_rw_lock_writer_lock(&connections_rwlock);
    g_hash_table_replace(connections, GINT_TO_POINTER(socket_id), data);
    g_static_rw_lock_writer_unlock(&connections_rwlock);
}

static void
remove_connection(int socket_id)
{
    NNTPConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        return;
    }

    if (data->ssl) {
        SSL_free(data->ssl);
    }

    g_static_rw_lock_writer_lock(&connections_rwlock);
    g_hash_table_remove(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_writer_unlock(&connections_rwlock);
}

static int
perform_read(int socket_id, void *buf, int len)
{
    NNTPConnectionData *data;
    fd_set recv_fds;
    struct timeval tv;
    int retval;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
#ifdef WIN32
        WSASetLastError(WSAEBADF);
#else
        errno = EBADF;
#endif
        return -1;
    }

    // Is there any data ready?
    FD_ZERO(&recv_fds);
    FD_SET(socket_id, &recv_fds);

    // Set read timeout to 30 seconds
    tv.tv_sec = 30;
    tv.tv_usec = 0;

    retval = select(socket_id + 1, &recv_fds, NULL, NULL, &tv);
    if (retval <= 0) {
#ifdef WIN32
        WSASetLastError(WSAETIMEDOUT);
#else
        errno = ETIMEDOUT;
#endif
        return -1;
    }

    if (data->use_ssl) {
        int ret = SSL_read(data->ssl, buf, len);
        switch(SSL_get_error(data->ssl, ret)) {
            case SSL_ERROR_NONE:
            case SSL_ERROR_ZERO_RETURN:
                errno = 0;
                break;

            case SSL_ERROR_WANT_READ:
                errno = EAGAIN;
                break;

            case SSL_ERROR_SSL:
#ifdef WIN32
                errno = ENOSYS;
#else
                errno = EPROTO;
#endif
                break;

            default:
                break;
        }

        return ret;
    } else {
        return recv(socket_id, buf, len, 0);
    }
}

static int
perform_write(int socket_id, const void *buf, int len)
{
    NNTPConnectionData *data;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    data = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    if (!data) {
        errno = EBADF;
        return -1;
    }

    if (data->use_ssl) {
        return SSL_write(data->ssl, buf, len);
    } else {
        return send(socket_id, buf, len, 0);
    }
}

static void
strip_newline(char *line)
{
    if (line[strlen(line) - 1]  == '\n') {
        line[strlen(line) - 1] = '\0';
    }

    if (line[strlen(line) - 1]  == '\r') {
        line[strlen(line) - 1] = '\0';
    }
}

#ifdef WIN32
static char *
WSAGetStrError(int err)
{
    LPVOID lpMsgBuf;
    char *ret;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        (DWORD) err,
        0,
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    ret = g_strdup(lpMsgBuf);
    strip_newline(ret);

    LocalFree(lpMsgBuf);

    return ret;
}
#endif

static void
nntp_disconnect(int conn_id, NNTPDisconnectType errcode, const char *errmsg, gboolean generate_errmsg_from_errno, gpointer data)
{
    const char *msg = NULL;

#ifdef WIN32
    if (generate_errmsg_from_errno) {
        msg = WSAGetStrError(WSAGetLastError());
    } else {
        msg = errmsg;
    }
#else
    if (generate_errmsg_from_errno) {
        msg = strerror(errno);
    } else {
        msg = errmsg;
    }
#endif

#ifdef WIN32
    closesocket(conn_id);
#else
    close(conn_id);
#endif

    remove_connection(conn_id);

    imported_funcs.notify_disconnect(errcode, msg, data);

#ifdef WIN32
    if (generate_errmsg_from_errno) {
        g_free((char*) msg);
    }
#endif
}

static char *
read_line(int socket_id, gboolean dont_buffer, gpointer data)
{
    int i;
    int len;
    char *line;
    time_t stamp;
    NNTPConnectionData *sock;

    g_static_rw_lock_reader_lock(&connections_rwlock);
    sock = g_hash_table_lookup(connections, GINT_TO_POINTER(socket_id));
    g_static_rw_lock_reader_unlock(&connections_rwlock);

    g_assert(sock);

    memset(sock->recv_buf + sock->recv_buf_len, 0, sizeof(sock->recv_buf) - sock->recv_buf_len);

    stamp = time(NULL);

    do {
        // Check is there is a newline in the buffer
        for (i = 0; i < sock->recv_buf_len; i++) {
            if (sock->recv_buf[i] == '\n') {
                // Newline found, split the buffer into a line and process it
                line = g_strndup(sock->recv_buf, i);

                memmove(sock->recv_buf, sock->recv_buf + i + 1, sock->recv_buf_len - i - 1);
                sock->recv_buf_len -= i + 1;
                memset(sock->recv_buf + sock->recv_buf_len, 0, sizeof(sock->recv_buf) - sock->recv_buf_len);

                return line;
            }
        }

        // No newline found, try to read new data from the socket
        do {
            if (dont_buffer) {      // Read only 1 byte until we've found a newline to avoid extra buffering
                char data[2];
                len = perform_read(socket_id, &data, 1);
                if (len == 1) {
                    sock->recv_buf[sock->recv_buf_len] = data[0];
                }
            } else {
                len = perform_read(socket_id, sock->recv_buf + sock->recv_buf_len, sizeof(sock->recv_buf) - sock->recv_buf_len);
            }

#ifdef WIN32
            if (len == -1 && (errno == EAGAIN || errno == EINTR || WSAGetLastError() == WSAETIMEDOUT)) {
#else
            if (len == -1 && (errno == EAGAIN || errno == EINTR)) {
#endif
                g_usleep(G_USEC_PER_SEC / 10);
            } else {
                break;
            }
        } while (stamp + 10 > time(NULL));      // timeout after 10 seconds

        if (len == -1) {
            // Error occured
            nntp_disconnect(socket_id, DISCONNECT_READ_ERROR, NULL, TRUE, data);
            return NULL;
        }

        if (len == 0) {
            // EOF, disconnect
            nntp_disconnect(socket_id, DISCONNECT_READ_ERROR, NULL, TRUE, data);
            g_print(__FILE__":%i EOF detected\n", __LINE__);

            return NULL;
        }

        perform_traffic_monitoring(len, data);
        sock->recv_buf_len += len;
    } while (TRUE);

g_print(__FILE__":%i Read timeout\n", __LINE__);
    nntp_disconnect(socket_id, DISCONNECT_READ_TIMEOUT, _("Read timeout"), FALSE, data);

    return NULL;
}

static gboolean
write_line(int socket_id, const char *line, gpointer data)
{
    if (perform_write(socket_id, line, strlen(line)) == -1) {
        nntp_disconnect(socket_id, DISCONNECT_WRITE_ERROR, NULL, TRUE, data);

        return FALSE;
    }

    return TRUE;
}

static NNTPGrabErrCode
login(const char *username, const char *password, char **errmsg, int *conn_id, gpointer data)
{
    char *line;
    char *welcome_msg;

    welcome_msg = read_line(*conn_id, FALSE, data);
    if (!welcome_msg) return NNTP_ERROR_READ;

    // Strip the newline character
    strip_newline(welcome_msg);

    switch(get_status_code(welcome_msg)) {
        case 200:
        case 201:
            // Login OK
            break;

        default:
            // Login not OK..probably too many connections from this address
            if (errmsg) {
                *errmsg = g_strdup(welcome_msg);
            }
            g_free(welcome_msg);

            nntp_disconnect(*conn_id, DISCONNECT_TOO_MANY_CONNECTIONS, *errmsg, FALSE, data);

            return NNTP_ERROR_TOO_MANY_CONNECTIONS;
    }

    // Do we need to login using a username and password ?
    if (username && strlen(username) > 0 && password && strlen(password) > 0) {
        line = g_strdup_printf("AUTHINFO USER %s\r\n", username);
        if (!write_line(*conn_id, line, data)) {
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);
            return NNTP_ERROR_WRITE;
        }

        g_free(line);

        line = read_line(*conn_id, FALSE, data);
        if (!line) return NNTP_ERROR_READ;

        if (get_status_code(line) != 381) {
            // Invalid reply, disconnect
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);

            nntp_disconnect(*conn_id, DISCONNECT_LOGIN_FAILURE, *errmsg, FALSE, data);

            return NNTP_ERROR_LOGIN_FAILURE;
        }

        g_free(line);

        line = g_strdup_printf("AUTHINFO PASS %s\r\n", password);
        if (!write_line(*conn_id, line, data)) {
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);
            return NNTP_ERROR_WRITE;
        }

        g_free(line);

        line = read_line(*conn_id, FALSE, data);
        if (!line) {
            return NNTP_ERROR_READ;
        }

        if (get_status_code(line) != 281) {
            // Invalid reply, disconnect
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);

            nntp_disconnect(*conn_id, DISCONNECT_LOGIN_FAILURE, *errmsg, FALSE, data);

            return NNTP_ERROR_LOGIN_FAILURE;
        }

        g_free(line);
    }

    // Tell the usenet server we want to use the READER mode
    line = g_strdup("MODE READER\r\n");
    if (!write_line(*conn_id, line, data)) {
        if (errmsg) {
            strip_newline(line);
            *errmsg = g_strdup(line);
        }
        g_free(line);

        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(*conn_id, FALSE, data);
    if (!line) return NNTP_ERROR_READ;

    switch(get_status_code(line)) {
        case 200:
        case 201:
            g_free(line);
            break;

        case 480:
            // Some usenet servers use this to notify there are too many connections active
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);

            nntp_disconnect(*conn_id, DISCONNECT_TOO_MANY_CONNECTIONS, *errmsg, FALSE, data);

            return NNTP_ERROR_TOO_MANY_CONNECTIONS;

        default:
            // Invalid reply, disconnect
            if (errmsg) {
                strip_newline(line);
                *errmsg = g_strdup(line);
            }
            g_free(line);

            nntp_disconnect(*conn_id, DISCONNECT_INVALID_MSG, *errmsg, FALSE, data);

            return NNTP_ERROR_INVALID_MSG;
    }

    // Login complete, let's party!

    imported_funcs.notify_connected(welcome_msg, data);
    g_free(welcome_msg);

    return NNTP_ERROR_NONE;
}

static SSL *
prepare_ssl_connection(int conn_id, char **errmsg)
{
    SSL_CTX     *ctx;
    SSL         *ssl;
    SSL_METHOD  *meth;
    int          err;
#if 0
    char        *str;
    X509        *server_cert;
#endif
    static      GStaticMutex mutex = G_STATIC_MUTEX_INIT;

    g_static_mutex_lock(&mutex);

    SSLeay_add_ssl_algorithms();
    meth = SSLv23_client_method();
    SSL_load_error_strings();
    g_static_mutex_unlock(&mutex);

    ctx = SSL_CTX_new (meth);
    if (!ctx) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_CTX_new FAILED");
        }

        return NULL;
    }

    // Don't verify the certificate as several usenet providers (like Eweka)
    // use self-signed certificates
    SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);

    ssl = SSL_new (ctx);
    SSL_CTX_free(ctx);

    if (!ssl) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_new FAILED");
        }

        return NULL;
    }

    SSL_set_fd (ssl, conn_id);
    err = SSL_connect (ssl);
    if (err <= 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf("%s", ERR_error_string(ERR_get_error(), NULL));
        }

        return NULL;
    }

#if 0
    printf ("SSL connection using %s\n", SSL_get_cipher (ssl));
    server_cert = SSL_get_peer_certificate (ssl);
    if (server_cert) {
        printf ("Server certificate:\n");

        str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
        if (str) {
            printf ("\t subject: %s\n", str);
            OPENSSL_free (str);
        }

        str = X509_NAME_oneline (X509_get_issuer_name  (server_cert),0,0);
        if (str) {
            printf ("\t issuer: %s\n", str);
            OPENSSL_free (str);
        }

        /* We could do all sorts of certificate verification stuff here before
           deallocating the certificate. */

        X509_free (server_cert);
    }
#endif

    return ssl;
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_connect(const char *hostname, int port, const char *username, const char *password, gboolean use_ssl, char **errmsg, int *conn_id, gpointer data)
{
    struct addrinfo hints, *res, *ressave;
    int n;
    char *str_port;
    int last_errno;
    SSL *ssl;
#ifdef WIN32
    int tv;
    u_long ioctlArg;
#else
    struct timeval tv;
    int mode;
#endif

    g_assert(hostname);
    g_assert(port > 0);
    g_assert(port < 65535);

    *conn_id = -1;

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    str_port = g_strdup_printf("%i", port);
    n = getaddrinfo(hostname, str_port, &hints, &res);
    g_free(str_port);

    if (n != 0) {
        if (errmsg) {
            *errmsg = g_strdup(gai_strerror(n));
        }
        return NNTP_ERROR_HOST_NOT_FOUND;
    }

    ressave = res;
    last_errno = 0;

    while (res) {
        int ret;

        *conn_id = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((*conn_id == INVALID_SOCKET)) {
#else
        if ((*conn_id == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
#ifdef WIN32
            last_errno = WSAGetLastError();
#else
            last_errno = errno;
#endif
            res = res->ai_next;
            *conn_id = -1;

            if (errmsg) {
                if (*errmsg) {
                    g_free(*errmsg);
                }

#ifdef WIN32
                *errmsg = WSAGetStrError(last_errno);
#else
                *errmsg = g_strdup(strerror(last_errno));
#endif
            }

            continue;
        }

        // Set the connection timeout on the socket
#ifdef WIN32
        tv = 5;

        setsockopt(*conn_id, SOL_SOCKET, SO_RCVTIMEO, (char*) &tv, sizeof(tv));
        setsockopt(*conn_id, SOL_SOCKET, SO_SNDTIMEO, (char*) &tv, sizeof(tv));
#else
        tv.tv_sec = 5;
        tv.tv_usec = 0;

        setsockopt(*conn_id, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(*conn_id, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));
#endif

        // Try to connect to the server
        imported_funcs.notify_connecting(*conn_id, data);

        // Mark the socket as non-blocking
#ifdef WIN32
        ioctlArg = 1;
        ioctlsocket(*conn_id, FIONBIO, &ioctlArg);
#else
        mode = fcntl(*conn_id, F_GETFL, 0);
        fcntl(*conn_id, F_SETFL, mode | O_NONBLOCK);
#endif

        ret = connect(*conn_id, res->ai_addr, (int) res->ai_addrlen);
        if (ret == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
#ifndef WIN32
        } else if (errno == EINPROGRESS) {
            // Wait for at most 5 seconds
            struct sockaddr name;
            socklen_t len = sizeof(struct addrinfo);

            time_t now = time(NULL);
            do {
                // According to http://cr.yp.to/docs/connect.html there are various
                // ways to detect if a socket is connected. The getpeername() method
                // should be the most portable one
                if (getpeername(*conn_id, &name, &len) == 0) {
                    errno = 0;
                    break;
                }

                g_usleep(G_USEC_PER_SEC / 10);
            } while (now + 5 > time(NULL));

            if (errno == 0) {
                last_errno = 0;
                break;
            }
#else
        } else if (WSAGetLastError() == WSAEWOULDBLOCK) {
            // And Windows again has different behaviour....
            struct timeval tv;
            fd_set send_fds;
            int len;

            FD_ZERO(&send_fds);
            FD_SET(*conn_id, &send_fds);

            tv.tv_sec = 5;
            tv.tv_usec = 0;

            len = select(*conn_id + 1, NULL, &send_fds, NULL, &tv);
            if (len <= 0) {
                // Read timeout or some other error
                if (len == 0) {
                    errno = WSAETIMEDOUT;
                }
            } else {
                // Connection is made
                last_errno = 0;
                break;
            }
#endif
        }

        // Connection could not be made, save the errno
#ifndef WIN32
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else
#endif
        {
            last_errno = errno;
        }

        if (errmsg) {
            if (*errmsg) {
                g_free(*errmsg);
            }

#ifdef WIN32
            *errmsg = WSAGetStrError(last_errno);
#else
            *errmsg = g_strdup(strerror(last_errno));
#endif
        }

#if WIN32
        if (last_errno == WSAETIMEDOUT) {
#else
        if (last_errno == ETIMEDOUT) {
#endif
            imported_funcs.notify_disconnect(DISCONNECT_CONNECT_TIMEOUT, (errmsg) ? *errmsg : NULL, data);
        } else {
            imported_funcs.notify_disconnect(DISCONNECT_CONNECTION_REFUSED, (errmsg) ? *errmsg : NULL, data);
        }

#ifdef WIN32
        closesocket(*conn_id);
#else
        close(*conn_id);
#endif

        *conn_id = -1;

        res = res->ai_next;
    }

    if (*conn_id == -1) {
        freeaddrinfo(ressave);

        switch (last_errno) {
#if WIN32
            case WSAETIMEDOUT:
#else
            case ETIMEDOUT:
#endif
                return NNTP_ERROR_CONNECTION_TIMEOUT;

            default:
                return NNTP_ERROR_CONNECTION_REFUSED;
        };
    }

    // Mark the socket as blocking
#ifdef WIN32
        ioctlArg = 0;
        ioctlsocket(*conn_id, FIONBIO, &ioctlArg);
#else
        mode = fcntl(*conn_id, F_GETFL, 0);
        fcntl(*conn_id, F_SETFL, mode  ^ O_NONBLOCK);
#endif

    freeaddrinfo(ressave);

    if (use_ssl) {
        if ((ssl = prepare_ssl_connection(*conn_id, errmsg)) == NULL) {
            nntp_disconnect(*conn_id, DISCONNECT_ERROR_SSL_INITIALISE, *errmsg, FALSE, data);
            return NNTP_ERROR_SSL_INITIALISE;
        }
    } else {
        ssl = NULL;
    }

    add_connection(*conn_id, use_ssl, ssl);

    return login(username, password, errmsg, conn_id, data);
}

void
nntpgrab_plugin_nntp_disconnect(int conn_id, gpointer data)
{
    nntp_disconnect(conn_id, DISCONNECT_NORMAL, NULL, FALSE, data);
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_get_article_by_message_id(int conn_id, const char *message_id, const char **headers, const char *filename, gpointer cb_data)
{
    FILE *file_tmp;
    FILE *file_dest;
    char *line;
    char *ptr;
    int fp;
    int length;
    int total_length = 0;
    GMappedFile *file;
    GError *err = NULL;
    struct timeval stamp;
    struct timeval prev_stamp;
    char write_buffer[128 * 1024];      // Save the data for every 128KB data received
    int write_buffer_len;
    int i;
    char buf[65535];
#ifdef WIN32
    char *filename_tmp;
#endif

    if (headers) {
        g_error(__FILE__":%i fetching of headers not yet implemented!\n", __LINE__);
    }

    imported_funcs.notify_download_begin(cb_data);

    line = g_strdup_printf("BODY %s\r\n", message_id);
    if (!write_line(conn_id, line, cb_data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(conn_id, TRUE, cb_data);
    if (!line) {
        return NNTP_ERROR_READ;
    }

    switch (get_status_code(line)) {
        case 222:            // article retrieved - body follows
            break;

        case 430:            // no such article found
            g_free(line);
            return NNTP_ERROR_PART_NOT_AVAILABLE;

        case 220:            // article retrieved - head and body follow
        case 221:            // article retrieved - head follows
        case 223:            // article retrieved - request text separately
        case 412:            // no newsgroup has been selected
        case 420:            // no current article has been selected
        case 423:            // no such article number in this group
        default:
            nntp_disconnect(conn_id, DISCONNECT_INVALID_MSG, line, FALSE, cb_data);
g_print(__FILE__":%i Invalid message received: %s\n", __LINE__, line);
            g_free(line);

            return NNTP_ERROR_INVALID_MSG;
    }

    g_free(line);

#ifdef WIN32
    fp = open(filename, O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, S_IRUSR | S_IWUSR);
#else
    fp = open(filename, O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR);
#endif

    gettimeofday(&stamp, NULL);
    prev_stamp = stamp;

    // Use a write buffer so we only need to perform a write every once in a while
    memset(&write_buffer, 0, sizeof(write_buffer));
    write_buffer_len = 0;

    while (TRUE) {
        char *end_of_data;
        fd_set rfds;
        struct timeval tv;
        int retval;
        int length_tmp;

        memset(&buf, 0, sizeof(buf));
        length = 0;
        while (length < 4) {
            // Guard for read timeouts
            FD_ZERO(&rfds);
            FD_SET(conn_id, &rfds);

            // read timeout of 30 seconds
            tv.tv_sec = 30;
            tv.tv_usec = 0;

            retval = select(conn_id + 1, &rfds, NULL, NULL, &tv);
            if (retval <= 0) {
                // Read timeout
                nntp_disconnect(conn_id, DISCONNECT_READ_TIMEOUT, _("Read timeout"), FALSE, cb_data);

                close(fp);
                unlink(filename);

                return NNTP_ERROR_READ;
            }

            // There should be data waiting now
            length_tmp = perform_read(conn_id, buf + length, sizeof(buf) - length);
            if (length_tmp <= 0) {      // The return value of the read call can be 0 when the connection has been gracefully closed
                nntp_disconnect(conn_id, DISCONNECT_READ_ERROR, NULL, TRUE, cb_data);

                close(fp);
                unlink(filename);

                return NNTP_ERROR_READ;
            }

            perform_traffic_monitoring(length_tmp, cb_data);

            length += length_tmp;
        }

        // Check if we are at the end of the message
        end_of_data = buf + length - 4;
        if (!strcmp(end_of_data, "\n.\r\n") || !strcmp(end_of_data + 1, "\n.\n")) {
            // Only keep the first part of the chunk
            if (!strcmp(end_of_data, "\n.\r\n")) {
                length -= 3;
            } else {
                length -= 2;
            }

            total_length += length;

            // Flush the write buffer if necessary
            if (write_buffer_len > 0) {
                if (write(fp, write_buffer, write_buffer_len) == -1) {
                    close(fp);
                    unlink(filename);

                    return NNTP_ERROR_UNABLE_TO_SAVE_PART;
                }
            }

            // Write the last piece of data
            if (write(fp, buf, length) == -1) {
                close(fp);
                unlink(filename);

                return NNTP_ERROR_UNABLE_TO_SAVE_PART;
            }

            // Break out of the loop as we are at the end of the message
            break;
        }

        // Do we need to flush the write buffer?
        if (write_buffer_len + length > sizeof(write_buffer)) {
            // Data doesn't fit in the buffer anymore, flush the buffer
            // Try to save the data to the disk
            if (write(fp, write_buffer, write_buffer_len) == -1) {
                // Disconnect the NNTP connection because we are now half-way transferring a part
                nntp_disconnect(conn_id, DISCONNECT_NORMAL, NULL, FALSE, cb_data);

                close(fp);
                unlink(filename);

                return NNTP_ERROR_UNABLE_TO_SAVE_PART;
            }

            memset(&write_buffer, 0, sizeof(write_buffer));
            write_buffer_len = 0;
        }

        // Add the chunk of data to the write buffer
        g_assert(write_buffer_len + length <= sizeof(write_buffer));

        memcpy(write_buffer + write_buffer_len, buf, length);
        write_buffer_len += length;

        total_length += length;

        // Send a progress update every 1/10th second
        gettimeofday(&stamp, NULL);
        timersub(&stamp, &prev_stamp, &tv);

        if (tv.tv_sec > 0 || tv.tv_usec >= 100000) {
            imported_funcs.update_progress(total_length, cb_data);
            prev_stamp = stamp;
        }
    }

    imported_funcs.update_progress(total_length, cb_data);

    close(fp);

    file = g_mapped_file_new(filename, TRUE, &err);
    if (!file) {
        g_print(__FILE__ ":%i g_mapped_file_new FAILED: %s\n", __LINE__, err->message);
        g_error_free(err);
        return NNTP_ERROR_UNABLE_TO_SAVE_PART;
    }

    // Do we need to un-escape any double-dots ?
    ptr = g_mapped_file_get_contents(file);
    length = g_mapped_file_get_length(file);
    for (i = 0; i < length - 3; i++) {
        if (ptr[i] == '\n' && ptr[i + 1] == '.' && ptr[i + 2] == '.') {
            memmove(ptr + i + 1, ptr + i + 2, length - i - 2);
            length -= 1;
        }
    }

    // Some platforms (like Windows) don't allow changing files which are already open
    // To work around this, use a temporary file
#ifdef WIN32
    // tmpfile() is broken on Vista....
    filename_tmp = _tempnam(g_get_user_cache_dir(), NULL);
    if (!filename_tmp) {
        g_print(__FILE__ ":%i tmpnam FAILED: %s\n", __LINE__, strerror(errno));
        g_mapped_file_free(file);
        return NNTP_ERROR_UNABLE_TO_SAVE_PART;
    }
    file_tmp = fopen(filename_tmp, "wb+");
#else
    file_tmp = tmpfile();
#endif
    if (!file_tmp) {
        g_print(__FILE__ ":%i tmpfile() FAILED: %s\n", __LINE__, strerror(errno));
        g_mapped_file_free(file);
        return NNTP_ERROR_UNABLE_TO_SAVE_PART;
    }

    if (fwrite(ptr, length, 1, file_tmp) != 1) {
        g_print(__FILE__ ":%i fwrite FAILED: %s\n", __LINE__, strerror(errno));
        g_mapped_file_free(file);
        fclose(file_tmp);
        return NNTP_ERROR_UNABLE_TO_SAVE_PART;
    }

    g_mapped_file_free(file);

    // Write the contents of the temporary file back to the wanted filename
    file_dest = fopen(filename, "wb");
    if (!file_dest) {
        g_print(__FILE__ ":%i fopen() FAILED: %s\n", __LINE__, strerror(errno));
        fclose(file_tmp);
        return NNTP_ERROR_UNABLE_TO_SAVE_PART;
    }

    rewind(file_tmp);
   
    while (!feof(file_tmp)) {
        length = fread(buf, 1, sizeof(buf), file_tmp);
        if (length > 0) {
            if (fwrite(buf, 1, length, file_dest) != length) {
                g_print("fwrite FAILED: %s\n", strerror(errno));
                fclose(file_tmp);
                fclose(file_dest);
                return NNTP_ERROR_UNABLE_TO_SAVE_PART;
            }
        }
    }

    fclose(file_tmp);
    fclose(file_dest);

#ifdef WIN32
    g_unlink(filename_tmp);
    free(filename_tmp);
#endif
 
    return NNTP_ERROR_NONE;
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_group(int conn_id, const char *newsgroup, int *num_articles, gint64 *start_article, gint64 *end_article, gpointer data)
{
    char *line;

    g_assert(newsgroup);
    g_assert(num_articles);
    g_assert(start_article);
    g_assert(end_article);

    line = g_strdup_printf("GROUP %s\r\n", newsgroup);
    if (!write_line(conn_id, line, data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }
    g_free(line);

    line = read_line(conn_id, FALSE, data);
    if (!line) return NNTP_ERROR_READ;

    switch (get_status_code(line)) {
        case 211:
            /*  211 n f l s group selected
                   (n = estimated number of articles in group,
                    f = first article number in the group,
                    l = last article number in the group,
                    s = name of the group.)
            */
#if SIZEOF_LONG == 8
            sscanf(line, "%*i %i %li %li %*s", num_articles, start_article, end_article);
#else
            sscanf(line, "%*i %i %lli %lli %*s", num_articles, start_article, end_article);
#endif
            g_free(line);
            return NNTP_ERROR_NONE;

        case 411:
            // No such group
            g_free(line);
            return NNTP_ERROR_NO_SUCH_GROUP;

        default:
            // Invalid response
            nntp_disconnect(conn_id, DISCONNECT_INVALID_MSG, line, FALSE, data);
            g_free(line);

            return NNTP_ERROR_INVALID_MSG;
    }
}

NNTPGrabErrCode
nntpgrab_plugin_nntp_xover(int conn_id, gint64 start_range, gint64 end_range, gpointer data)
{
    char *line;
    char **parts;
    struct tm tm;
    time_t post_date;

    g_assert(start_range >= 1);
    g_assert(end_range >= -1);

    // TODO: found out the order of the fields using the command 'LIST OVERVIEW.FMT'
    // On NewsZilla, this command returns this list:
    // LIST OVERVIEW.FMT
    // 215 Order of fields in overview database.
    // Subject:
    // From:
    // Date:
    // Message-ID:
    // References:
    // Bytes:
    // Lines:
    // Xref:full
    // .
    //
    // The word 'full' has an additional meaning according to RFC2980:
    //
    // Please note that if the header has the word "full" (without quotes)
    // after the colon, the header's name is prepended to its field in the
    // output returned by the server.

    if (end_range == -1) {
#if SIZEOF_LONG == 8
        line = g_strdup_printf("XOVER %li-\r\n", start_range);
#else
        line = g_strdup_printf("XOVER %lli-\r\n", start_range);
#endif
    } else {
#if SIZEOF_LONG == 8
        line = g_strdup_printf("XOVER %li-%li\r\n", start_range, end_range);
#else
        line = g_strdup_printf("XOVER %lli-%lli\r\n", start_range, end_range);
#endif
    }

    if (!write_line(conn_id, line, data)) {
        g_free(line);
        return NNTP_ERROR_WRITE;
    }

    g_free(line);

    line = read_line(conn_id, FALSE, data);
    if (!line) return NNTP_ERROR_READ;

    switch (get_status_code(line)) {
        case 224:
            // 224 Overview information follows
            g_free(line);
            break;

        default:
            // Invalid response
            nntp_disconnect(conn_id, DISCONNECT_INVALID_MSG, line, FALSE, data);
            g_free(line);

            return NNTP_ERROR_INVALID_MSG;
    }

    while (1) {
        line = read_line(conn_id, FALSE, data);
        if (!line) return NNTP_ERROR_READ;

        if (strncmp(line, ".", 1) == 0) {
            g_free(line);
            break;
        }

        parts = g_strsplit(line, "\t", 0);

        /* Prevent segfaults */
        if (!parts[0]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[1]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[2]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[3]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[4]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[5]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[6]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        } else if (!parts[7]) {
            g_strfreev(parts);
            g_free(line);
            continue;
        }

        memset(&tm, 0, sizeof(tm));
        if (strptime(parts[3], "%d %b %Y %H:%M:%S %Z", &tm)) {
            post_date = mktime(&tm);
        } else {
            memset(&tm, 0, sizeof(tm));
            if (strptime(parts[3], "%A, %d %b %Y %H:%M:%S %Z", &tm)) {
                post_date = mktime(&tm);
            } else {
                // Date could not be parsed
                post_date = 0;

                g_strfreev(parts);
                g_free(line);
                continue;
            }
        }

        imported_funcs.parse_header(atoi(parts[0]), parts[1], parts[2], post_date, parts[4], atoi(parts[6]), atoi(parts[7]), start_range, end_range, data);

        g_strfreev(parts);
        g_free(line);
    }

    return NNTP_ERROR_NONE;
}
