/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NNTPGRAB_PLUGIN_NNTP_H_
#define _NNTPGRAB_PLUGIN_NNTP_H_

#include "nntpgrab_plugin_base.h"

// These are all the functions that the plugin can call to communicate with the NNTPGrab core
typedef struct _plugin_nntp_imported_funcs {
    int version;

    void (*update_progress) (int bytes_downloaded, gpointer data);
    void (*update_traffic_monitor) (int bytes_received[10], time_t stamp);
    void (*notify_download_begin) (gpointer data);
    void (*notify_connecting) (int conn_id, gpointer data);
    void (*notify_connected) (const char *welcome_msg, gpointer data);
    void (*notify_disconnect) (NNTPDisconnectType disconnect_type, const char *errmsg, gpointer data);
    void (*parse_header) (int position, const char *subject, const char *poster, time_t postdate, const char *message_id, int size, int numlines, int start_range, int end_range, gpointer data);
} PluginNNTPImportedFuncs;

// The plugin needs to exported these methods :

/**
 * Initialize the plugin
 *
 * @arg     funcs   A structure containing all the callback functions the plugin can call to notify the NNTPGrab Core
 *
 * @return          TRUE when the initialization was successfull
 */
gboolean     nntpgrab_plugin_nntp_initialize(PluginNNTPImportedFuncs funcs);

/**
 * Return the API version of the plugin. This is used to verify that the plugin is compatible
 * with the interface the NNTPGrab Core expects
 *
 * @return  NNTPGRAB_PLUGIN_NNTP_API_VERSION
 */
int          nntpgrab_plugin_nntp_get_version(void);

/**
 * Function to check is the connection is still alive
 *
 * @arg     conn_id     The connection ID to check
 *
 * @return  TRUE when the connection is still alive
 */
gboolean     nntpgrab_plugin_nntp_get_is_connected(int conn_id);

/**
 * Connect to the specified usenet server with the given parameters
 *
 * @arg     hostname    The hostname (or IP address) of the usenet server to connect to
 * @arg     port        The port to connect to
 * @arg     username    A username which is required to login, can be NULL
 * @arg     password    A password which is required to login, can be NULL
 * @arg     use_ssl     Do we need to use NNTP over SSL?
 * @arg     errmsg      If an error occurs, a message can be saved here, can be NULL
 * @arg     conn_id     Pointer to a integer where the connection ID will be saved
 * @arg     data        Data to be used by callback functions
 *
 * @return  ERROR_NONE when the connection and login are successfull
 */
NNTPGrabErrCode  nntpgrab_plugin_nntp_connect(const char *hostname, int port, const char *username, const char *password, gboolean use_ssl, char **errmsg, int *conn_id, gpointer data);

/**
 * Disconnect from the connected server
 *
 * @arg     conn_id     The connection ID
 * @arg     data        Data to be used by callback functions
 *
 */
void             nntpgrab_plugin_nntp_disconnect(int conn_id, gpointer data);

/**
 * Get a article from the connected usenet server
 *
 * @arg     conn_id     The connection ID
 * @arg     message_id  The message-id of the article requested
 * @arg     headers     Pointer to catch all the headers, can be NULL
 * @arg     filename    The filename of the file where the message data needs to be saved
 * @arg     cb_data     Data to be used by callback functions
 *
 * @return  ERROR_NONE on success
 */
NNTPGrabErrCode  nntpgrab_plugin_nntp_get_article_by_message_id(int conn_id, const char *message_id, const char **headers, const char *filename, gpointer cb_data);

/**
 * Set the active newsgroup
 *
 * @arg     conn_id         The connection ID
 * @arg     newsgroup       The name of the newsgroup which need to be selected
 * @arg     num_articles    Pointer to set the number of articles present in the given group
 * @arg     start_article   Pointer to set the first article number of the given group
 * @arg     end_article     Pointer to set the last article number of the given group
 * @arg     data            Data to be used by callback functions
 *
 */
NNTPGrabErrCode  nntpgrab_plugin_nntp_group(int conn_id, const char *newsgroup, int *num_articles, gint64 *start_article, gint64 *end_article, gpointer data);

/**
 * Retrieve a list of the available messages in active newsgroup
 *
 * @arg     conn_id     The connection ID
 * @arg     start_range The beginning of the range.
 * @arg     end_range   The end of the range. If -1, then all available messages are checked
 * @arg     data        Data to be used by callback functions
 *
 */
NNTPGrabErrCode  nntpgrab_plugin_nntp_xover(int conn_id, gint64 start_range, gint64 end_range, gpointer data);

#endif /* _NNTPGRAB_PLUGIN_NNTP_H_ */
