/*
    Copyright (C) 2005-2008  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#ifdef ENABLE_GUI
#include "nntpgrab_gui_base.h"
#endif

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>    /* freeaddrinfo for Win2k */
#else
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif /* WIN32 */

#include "nntpgrab.h"
#include "nntpgrab_automation.h"

#ifdef WIN32
#define CLOSE(x)    closesocket(x)
#else
#define CLOSE(x)    close(x)
#endif

static gboolean abort_flag = FALSE;
static GMainLoop *loop = NULL;
static GList *connections = NULL;
static NNTPGrabCore *core = NULL;

#ifdef ENABLE_GUI
static gboolean gui_enabled = FALSE;
#endif

// signals.c
void    connect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *send_buffer);
void    disconnect_signal_handlers(NNTPGrabCore *core, GAsyncQueue *send_buffer);

// commands.c
void    server_config_get_avail_servers(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_get_server_info(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_add_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_del_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_edit_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_get_opts(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_set_opts(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_config_get_folder_listing(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_start(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_stop(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_get_state(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_add_task_to_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_del_task_from_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_restart_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_save_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_foreach_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_move_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);
void    server_schedular_move_collection(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer);

typedef struct _connection {
    int socket_id;
    gboolean is_listener;
    gboolean is_command;
    char recv_buf[65536];      // It should be safe to assume that one line is less than 64KB
    int recv_buf_len;
    GAsyncQueue *send_buffer;
} Connection;

#ifdef ENABLE_GUI
static gboolean
update_connection_count_cb(gpointer data)
{
    int count = GPOINTER_TO_INT(data);
    GtkWidget *lblNumActiveConnections = nntpgrab_gui_base_get_widget("lblNumActiveConnections");
    char *msg = g_strdup_printf("%i", count);

    gtk_label_set_text(GTK_LABEL(lblNumActiveConnections), msg);

    g_free(msg);

    return FALSE;
}
#endif

static void
update_connection_count(int count)
{
#ifdef ENABLE_GUI
    g_idle_add(update_connection_count_cb, GINT_TO_POINTER(count));
#endif
}

static gboolean
process_command(const char *command, Connection *conn)
{
g_print("received command from socket %i: %s\n", conn->socket_id, command);
    if (!strncmp("QUIT", command, 4)) {
        CLOSE(conn->socket_id);
        conn->socket_id = -1;

#if 0
        g_static_rw_lock_writer_lock(&rwlock);
        connections = g_list_remove(connections, conn);
        g_free(conn);
        g_static_rw_lock_writer_unlock(&rwlock);
#endif

        return TRUE;
    }

    if (!strncmp("MODE LISTENER", command, 13)) {
        if (!conn->is_listener) {
            connect_signal_handlers(core, conn->send_buffer);
        }

        conn->is_listener = TRUE;
        conn->is_command = FALSE;

        g_async_queue_push(conn->send_buffer, g_strdup("OK\n"));

        return TRUE;
    }

    if (!strncmp("MODE COMMAND", command, 12)) {
        if (conn->is_listener) {
            disconnect_signal_handlers(core, conn->send_buffer);
        }

        conn->is_listener = FALSE;
        conn->is_command = TRUE;

        send(conn->socket_id, "OK\n", 3, 0);

        return TRUE;
    }

    if (!strncmp("KILL_SERVER", command, 11)) {
        abort_flag = TRUE;

        send(conn->socket_id, "OK\n", 3, 0);

        return TRUE;
    }

    if (!conn->is_command) {
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(conn->send_buffer, g_strdup("INVALID COMMAND\n"));
        return TRUE;
    }

    if (!strncmp("CONFIG_GET_AVAIL_SERVERS", command, 24)) {
        server_config_get_avail_servers(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_GET_SERVER_INFO", command, 22)) {
        server_config_get_server_info(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_ADD_SERVER", command, 17)) {
        server_config_add_server(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_DEL_SERVER", command, 17)) {
        server_config_del_server(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_EDIT_SERVER", command, 18)) {
        server_config_edit_server(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_GET_OPTS", command, 15)) {
        server_config_get_opts(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_SET_OPTS", command, 15)) {
        server_config_set_opts(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("CONFIG_GET_FOLDER_LISTING", command, 25)) {
        server_config_get_folder_listing(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_START", command, 15)) {
        server_schedular_start(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_STOP", command, 14)) {
        server_schedular_stop(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_GET_STATE", command, 19)) {
        server_schedular_get_state(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_ADD_TASK_TO_QUEUE", command, 27)) {
        server_schedular_add_task_to_queue(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_DEL_TASK_FROM_QUEUE", command, 29)) {
        server_schedular_del_task_from_queue(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_RESTART_TASK", command, 22)) {
        server_schedular_restart_task(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_SAVE_QUEUE", command, 20)) {
        server_schedular_save_queue(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_FOREACH_TASK", command, 22)) {
        server_schedular_foreach_task(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_MOVE_TASK", command, 19)) {
        server_schedular_move_task(core, command, conn->send_buffer);
        return TRUE;
    }

    if (!strncmp("SCHEDULAR_MOVE_COLLECTION", command, 25)) {
        server_schedular_move_collection(core, command, conn->send_buffer);
        return TRUE;
    }

    return FALSE;
}

static gboolean
try_to_read_line(Connection *sock)
{
    int i;
    int len;
    char *line;
    time_t stamp;

    g_assert(sock);

    memset(sock->recv_buf + sock->recv_buf_len, 0, sizeof(sock->recv_buf) - sock->recv_buf_len);

    stamp = time(NULL);

    do {
        struct timeval tv;
        fd_set recv_fds;

        // Check if there is a newline in the buffer
        for (i = 0; i < sock->recv_buf_len; i++) {
            if (sock->recv_buf[i] == '\n') {
                // Newline found, split the buffer into a line and process it
                line = g_strndup(sock->recv_buf, i);

                // Strip any possible \r characters
                if (line[strlen(line) - 1]  == '\r') {
                    line[strlen(line) - 1] = '\0';
                }

                if (!process_command(line, sock)) {
                    g_async_queue_push(sock->send_buffer, g_strdup("INVALID COMMAND\n"));
                }

                g_free(line);

                memmove(sock->recv_buf, sock->recv_buf + i + 1, sock->recv_buf_len - i - 1);
                sock->recv_buf_len -= i + 1;

                return TRUE;
            }
        }

        // No newline found, try to read new data from the socket
        do {
            FD_ZERO(&recv_fds);
            FD_SET(sock->socket_id, &recv_fds);

            tv.tv_sec = 5;  // Timeout after 5 seconds
            tv.tv_usec = 0;

            len = select(sock->socket_id + 1, &recv_fds, NULL, NULL, &tv);
            if (len == 0) {
                // Read timeout
                return FALSE;
            } else if (len == 1) {
                gboolean beenhere = FALSE;

                do {
                    if (beenhere) {
                        g_usleep(G_USEC_PER_SEC / 10);
                    } else {
                        beenhere = TRUE;
                    }

                    len = recv(sock->socket_id, sock->recv_buf + sock->recv_buf_len, sizeof(sock->recv_buf) - sock->recv_buf_len, 0);
                } while ((errno == EAGAIN) && stamp + 5 > time(NULL));
            }
        } while (len == -1 && errno == EINTR && stamp + 5 > time(NULL));

        if (len == -1) {
            if (errno == EAGAIN) {
                // Read timeout
                return FALSE;
            }

            // Error occured
g_print(__FILE__":%i Error occured: %s\n", __LINE__, strerror(errno));

            return FALSE;
        }

        if (len == 0) {
            // EOF, disconnect
g_print(__FILE__":%i EOF detected\n", __LINE__);

            return FALSE;
        }

        sock->recv_buf_len += len;
    } while (TRUE);

    return FALSE;
}

typedef struct _listen_socket {
    int socket_id;
    int ai_family;
} ListenSocket;

static gpointer
listen_thread(gpointer data)
{
    struct addrinfo hints, *res, *ressave;
    int n;
    char *port;
    int socket_id;
    struct timeval tv;
    fd_set recv_fds, send_fds;
    int retval;
    char *welcome_msg;
    GList *list;
    int top_socket_id;
    GList *listen_sockets = NULL;
    int errno_saved = 0;

    welcome_msg = g_strdup_printf("NNTPGrab Server - API version %i\n", NNTPGRAB_API_VERSION);

    memset(&hints, 0, sizeof(struct addrinfo));

    /*
       AI_PASSIVE flag: the resulting address is used to bind
       to a socket for accepting incoming connections.
       So, when the hostname==NULL, getaddrinfo function will
       return one entry per allowed protocol family containing
       the unspecified address for that family.
    */

    hints.ai_flags    = AI_PASSIVE;
    hints.ai_family   = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    port = g_strdup_printf("%i", 5423);     // FIXME: Needs to be configurable
    n = getaddrinfo(NULL, port, &hints, &res);
    g_free(port);

    if (n < 0) {
        fprintf(stderr,
                "getaddrinfo error:: [%s]\n",
#ifdef WIN32
                WSAGetLastError());
#else
                gai_strerror(n));
#endif
        return NULL;
    }

    ressave=res;

    /*
       Try open socket with each address getaddrinfo returned,
       until getting a valid listening socket.
    */
    socket_id = -1;
    while (res) {
        ListenSocket *sock;

        socket_id = socket(res->ai_family,
                           res->ai_socktype,
                           res->ai_protocol);

        if (socket_id < 0) {
            errno_saved = errno;
            res = res->ai_next;
            continue;
        }

        if (bind(socket_id, res->ai_addr, (int) res->ai_addrlen) == -1) {
            errno_saved = errno;
            CLOSE(socket_id);
            res = res->ai_next;
            continue;
        }

        if (listen(socket_id, 5) == -1) {
            errno_saved = errno;
            CLOSE(socket_id);
        }

        sock = g_slice_new0(ListenSocket);
        sock->socket_id = socket_id;
        sock->ai_family = res->ai_family;

        listen_sockets = g_list_append(listen_sockets, sock);
        res = res->ai_next;
    }

    if (listen_sockets == NULL) {
        freeaddrinfo(ressave);
#ifdef ENABLE_GUI
        if (gui_enabled) {
            GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");
            char *msg = g_strdup_printf("Unable to open socket: %s", strerror(errno_saved));
            g_critical(msg);
            g_free(msg);
            g_idle_add((GSourceFunc) gtk_widget_hide, windowMain);
            g_idle_add((GSourceFunc) gtk_main_quit, NULL);
        }
#endif
        fprintf(stderr, "socket error:: could not open socket: %s\n", strerror(errno_saved));
#ifndef ENABLE_GUI
        exit(-1);
#endif
        return NULL;
    }

    g_print("NNTPGrab Server up and running on %i sockets\n\n", g_list_length(listen_sockets));

    while (!abort_flag) {
        FD_ZERO(&recv_fds);
        FD_ZERO(&send_fds);

        list = listen_sockets;
        top_socket_id = 0;
        while (list) {
            ListenSocket *sock = list->data;
            FD_SET(sock->socket_id, &recv_fds);

            if (top_socket_id < sock->socket_id) {
                top_socket_id = sock->socket_id;
            }

            list = g_list_next(list);
        }

        list = connections;
        while (list) {
            Connection *conn = (Connection*) list->data;

            // Is this socket still alive ?
            if (conn->socket_id < 0) {
g_print("dead socket detected\n");
                list = g_list_next(list);
                connections = g_list_remove(connections, conn);
                update_connection_count(g_list_length(connections));
                g_slice_free(Connection, conn);
                continue;
            }

            FD_SET(conn->socket_id, &recv_fds);

            if (g_async_queue_length(conn->send_buffer) > 0) {
                FD_SET(conn->socket_id, &send_fds);
            }

            if (conn->socket_id > top_socket_id) {
                top_socket_id = conn->socket_id;
            }

            list = g_list_next(list);
        }

        tv.tv_sec = 0;
        tv.tv_usec = G_USEC_PER_SEC / 10;

        retval = select((int) (top_socket_id + 1), &recv_fds, &send_fds, NULL, &tv);
        if (retval == -1) {
            perror("select");
            exit(-1);
        } else if (retval > 0) {
            // Has anything happened on one of the listener sockets ?
            list = listen_sockets;
            while (list) {
                ListenSocket *sock = list->data;

                if (FD_ISSET(sock->socket_id, &recv_fds)) {
                    struct sockaddr_in clientname;
                    struct sockaddr_in6 clientname6;
                    struct sockaddr *ptr;
                    int new_fd;
                    Connection *conn = g_slice_new0(Connection);
#ifdef WIN32
                    int size;
#else
                    socklen_t size;
#endif

                    if (sock->ai_family == AF_INET6) {
                        ptr = (struct sockaddr*) &clientname6;
                        size = sizeof(struct sockaddr_in6);
                    } else {
                        ptr = (struct sockaddr*) &clientname;
                        size = sizeof(struct sockaddr_in);
                    }

                    new_fd = accept(sock->socket_id, ptr, &size);
                    if (new_fd < 0) {
#ifdef WIN32
                        g_error("Error occured during accept. WSAGetLastError() = %i\n", WSAGetLastError());
#else
                        perror("accept");
#endif
                        exit(EXIT_FAILURE);
                    }

                    conn->socket_id = new_fd;
                    conn->send_buffer = g_async_queue_new();
                    g_async_queue_push(conn->send_buffer, g_strdup(welcome_msg));

                    connections = g_list_append(connections, conn);
                    update_connection_count(g_list_length(connections));

                    g_print("New client connected\n");
                }

                list = g_list_next(list);
            }

            list = connections;
            while (list) {
                Connection *conn = (Connection*) list->data;
                int socket_id_saved = conn->socket_id;

                if (FD_ISSET(conn->socket_id, &recv_fds)) {
                    if (!try_to_read_line(conn) || conn->socket_id < 0) {
                        // Error occured, disconnect
                        gpointer data;

                        g_print("socket %i disconnected\n", socket_id_saved);

                        if (conn->socket_id >= 0) {
                            CLOSE(conn->socket_id);
                        }

                        while ((data = g_async_queue_try_pop(conn->send_buffer))) {
                            g_free(data);
                        }

                        if (conn->is_listener) {
                            disconnect_signal_handlers(core, conn->send_buffer);
                        }

                        g_async_queue_unref(conn->send_buffer);
                        g_slice_free(Connection, conn);

                        list = g_list_next(list);
                        connections = g_list_remove(connections, conn);
                        update_connection_count(g_list_length(connections));

                        continue;
                    }
                }

                if (FD_ISSET(conn->socket_id, &send_fds)) {
                    char *cmd;

                    while ((cmd = g_async_queue_try_pop(conn->send_buffer))) {
#if defined(WIN32) || defined(DARWIN)
                        send(conn->socket_id, cmd, strlen(cmd), 0);
#else
                        send(conn->socket_id, cmd, strlen(cmd), MSG_NOSIGNAL);
#endif
                        g_free(cmd);
                    }
                }

                list = g_list_next(list);
            }
        }
    }

    freeaddrinfo(ressave);

    // Kill all the active connections
    list = listen_sockets;
    while (list) {
        socket_id = GPOINTER_TO_INT(list->data);
#ifdef WIN32
        shutdown(socket_id, SD_BOTH);
#else
        shutdown(socket_id, SHUT_RDWR);
#endif
        CLOSE(socket_id);

        list = g_list_next(list);
    }
    g_list_free(listen_sockets);

    list = connections;
    while (list) {
        Connection *conn = (Connection*) list->data;
        gpointer data;

        while ((data = g_async_queue_try_pop(conn->send_buffer))) {
            g_free(data);
        }

        if (conn->is_listener) {
            disconnect_signal_handlers(core, conn->send_buffer);
        }

        g_async_queue_unref(conn->send_buffer);

        CLOSE(conn->socket_id);
        g_slice_free(Connection, conn);

        list = g_list_next(list);
    }
    g_list_free(connections);

    g_free(welcome_msg);

    g_main_loop_quit(loop);

    update_connection_count(0);

    return NULL;
}

#if GLIB_CHECK_VERSION(2,16,0)
static void
on_config_changed(NNTPGrabCore *core, gpointer data)
{
    static ConfigOpts prev_opts;
    static gboolean been_here = FALSE;
    ConfigOpts opts;
    char *errmsg = NULL;

    if (!been_here) {
        been_here = TRUE;
        memset(&prev_opts, 0, sizeof(ConfigOpts));
    }

    opts = nntpgrab_config_get_opts(core);

    if (prev_opts.enable_auto_import != opts.enable_auto_import) {
        if (opts.enable_auto_import) {
            if (!nntpgrab_automation_enable_auto_import(core, opts.auto_import_directory, &errmsg)) {
                g_warning("%s\n", errmsg);
                g_free(errmsg);
            }
        } else {
            nntpgrab_automation_disable_auto_import();
        }
    }
}
#endif

#ifdef ENABLE_GUI

G_MODULE_EXPORT void
on_btnQuit_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowMain;

    windowMain = nntpgrab_gui_base_get_widget("windowMain");

    gtk_widget_hide(windowMain);
    gtk_main_quit();
}

static void
on_schedular_state_changed(NNTPGrabCore *core, SchedularState new_state, const char *reason)
{
    GtkWidget *btnPauseSchedular;
    GtkWidget *lblSchedularState;

    btnPauseSchedular = nntpgrab_gui_base_get_widget("btnPauseSchedular");
    lblSchedularState = nntpgrab_gui_base_get_widget("lblSchedularState");

    switch (new_state) {
        case SCHEDULAR_STATE_RUNNING:
            gtk_widget_set_sensitive(btnPauseSchedular, TRUE);
            gtk_button_set_label(GTK_BUTTON(btnPauseSchedular), _("Stop the schedular"));
            gtk_label_set_text(GTK_LABEL(lblSchedularState), _("Running"));
            break;

        case SCHEDULAR_STATE_STOPPING:
            gtk_widget_set_sensitive(btnPauseSchedular, FALSE);
            gtk_button_set_label(GTK_BUTTON(btnPauseSchedular), _("Stop the schedular"));
            gtk_label_set_text(GTK_LABEL(lblSchedularState), _("Stopping"));
            break;

        case SCHEDULAR_STATE_STOPPED:
        default:
            gtk_widget_set_sensitive(btnPauseSchedular, TRUE);
            gtk_button_set_label(GTK_BUTTON(btnPauseSchedular), _("Start the schedular"));
            gtk_label_set_text(GTK_LABEL(lblSchedularState), _("Stopped"));
            break;
    }
}

G_MODULE_EXPORT void
on_btnPauseSchedular_clicked(GtkWidget *caller, gpointer data)
{
    switch(nntpgrab_schedular_get_state(core)) {
        case SCHEDULAR_STATE_RUNNING:
            nntpgrab_schedular_stop(core, FALSE);
            break;

        case SCHEDULAR_STATE_STOPPED:
            nntpgrab_schedular_start(core);
            break;

        default:
            g_return_if_reached();
    }
}

static gboolean
print_handler2(gpointer data)
{
    char *message = (char*) data;
    static gboolean mark_initialized = FALSE;
    GtkTextIter iter;
    GtkWidget *textviewLog;
    GtkTextBuffer *buffer;
    GtkTextMark *mark;

    textviewLog = nntpgrab_gui_base_get_widget("textviewLog");
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textviewLog));

    gtk_text_buffer_get_end_iter(buffer, &iter);
    gtk_text_buffer_insert(buffer, &iter, message, -1);

    // Scroll to the end
    if (!mark_initialized) {
        mark_initialized = TRUE;
        gtk_text_buffer_create_mark (buffer, "scroll", &iter, TRUE);
    }

    mark = gtk_text_buffer_get_mark (buffer, "scroll");
    gtk_text_buffer_get_end_iter (buffer, &iter);

    // Move the iterator to the beginning of line, so we don't scroll in horizontal direction
    gtk_text_iter_set_line_offset (&iter, 0);

    gtk_text_buffer_move_mark (buffer, mark, &iter);
    gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(textviewLog), mark);

    g_free(message);

    return FALSE;
}

static void
nntpgrab_server_print_handler(const char *message)
{
    g_idle_add(print_handler2, g_strdup(message));
}
#endif

static void
on_par2_repair_message(NNTPGrabCore *core, const char *message, gpointer data)
{
    g_print("%s\n", message);
}

int
main(int argc, char **argv)
{
    GThread *listen;
    char *errors = NULL;
    char *warnings = NULL;
#ifdef WIN32
    WORD wVersionRequested;
    WSADATA wsaData;
    int ret;
#endif

    if (!g_thread_supported ()) g_thread_init (NULL);

#ifdef WIN32
    // exchndl.dll is part of Dr. MinGW. By loading this DLL, every time the
    // program crashes a .RPT file will be created containing the backtrace
    LoadLibraryA("exchndl.dll");

    // Initialise Winsock
    wVersionRequested = MAKEWORD(2, 2);

    ret = WSAStartup(wVersionRequested, &wsaData);
    if (ret != 0) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        g_error(_("Winsock could not be initialised"));
        return -1;
    }

    /* Confirm that the WinSock DLL supports 2.2.        */
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */

    if (LOBYTE( wsaData.wVersion ) != 2 ||
        HIBYTE( wsaData.wVersion ) != 2) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup();

        g_error(_("Winsock 2.2 or higher is required for this program"));

        return -1;
    }
#endif

#ifdef ENABLE_GUI
    // Is there an graphical environment running?
#if !defined(WIN32) && !defined(DARWIN)
    if (g_getenv("DISPLAY")) {
#else
    if (TRUE) {
#endif
        gtk_init(&argc, &argv);

        nntpgrab_gui_base_create_ui("nntpgrab_server");
        nntpgrab_gui_base_tray_create();
        g_set_print_handler(nntpgrab_server_print_handler);
        gui_enabled = TRUE;
    }
#endif

    g_print(_("Now initialising the NNTPGrab Core...\n"));
    core = nntpgrab_core_init(NNTPGRAB_API_VERSION, &errors, &warnings);
    if (!core) {
        g_critical(_("NNTPGrab initialisation FAILED: %s\n"), errors);
#ifdef ENABLE_GUI
        g_idle_add((GSourceFunc) gtk_main_quit, NULL);
        gtk_main();
#endif
        exit(1);
    }

#ifdef ENABLE_GUI
    if (gui_enabled) {
        GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");
        gtk_widget_show(windowMain);

        g_signal_connect(core, "schedular_state_changed", G_CALLBACK(on_schedular_state_changed), NULL);
    }
#endif

    nntpgrab_schedular_start(core);

    if (warnings) {
        g_print("Warning:\n%s\n", warnings);
        g_free(warnings);
    }

#if GLIB_CHECK_VERSION(2,16,0)
    g_signal_connect(core, "config_changed", G_CALLBACK(on_config_changed), NULL);
#endif

    g_signal_connect(core, "par2_repair_message", G_CALLBACK(on_par2_repair_message), NULL);

    listen = g_thread_create(listen_thread, NULL, FALSE, NULL);

#ifdef ENABLE_GUI
    if (gui_enabled) {
        gtk_main();
    } else {
#endif
        loop = g_main_loop_new(NULL, FALSE);

        g_main_loop_run(loop);

        g_main_loop_unref(loop);
        g_main_context_unref(g_main_context_default());
#ifdef ENABLE_GUI
    }
#endif

#ifdef ENABLE_GUI
#if GTK_CHECK_VERSION(2,10,0)
    if (gui_enabled && nntpgrab_gui_base_tray_icon_get_is_shown()) {
        nntpgrab_gui_base_tray_destroy();
    }
#endif
#endif

    nntpgrab_core_cleanup(core);

    return 0;
}
